import os
import logging
from decimal import Decimal
from datetime import datetime
from dotenv import load_dotenv
from sqlalchemy import Column, Integer, String, Unicode, Numeric, ForeignKey, DateTime, Boolean, Enum, event
from sqlalchemy.orm import relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import StringEncryptedType
import enum
from broker_clients import Base, session
# no delete
from broker_clients.binance_proxy.storage.models import Orders, Fills, SymbolMetadata
from broker_clients.quantfury_proxy.storage.models import QuantfuryPositionType, QuantfuryPositions, QuantfuryOrders

load_dotenv()
logger = logging.getLogger(__name__)
cypher_key = os.getenv("CYPHER_KEY")


class TradeInTrend(enum.Enum):
    none = 'none'
    first_upper_scale = 'first_upper_scale'
    second_upper_scale = 'second_upper_scale'
    both_upper_scales = 'both_upper_scales'
    any_upper_scales = 'any_upper_scales'
    second_any_upper_scales = 'second_any_upper_scales'


class TradeStatus(enum.Enum):
    hold = 'hold'
    open = 'open'
    close = 'close'
    cancel = 'cancel'
    target = 'target'


class Trades(Base):
    __tablename__ = 'trades'

    trade_id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(Unicode, nullable=False)
    symbol = Column(String, nullable=False)
    time_frame = Column(String, nullable=False)

    entry_date = Column(DateTime, nullable=False)
    amount = Column(Numeric, nullable=True)  # must be same on Orders.quote_order_quantity
    stop_loss = Column(Numeric, nullable=True)  # must be same on Orders.stop_price
    buy_price = Column(Numeric, nullable=True)  # must be same on Orders.price
    sell_price = Column(Numeric, nullable=True)  # must be same on Orders.price
    profit = Column(Numeric, nullable=True)  # must be same on Orders.price
    side = Column(String, nullable=False)
    low_moment_price = Column(Numeric, default=0, nullable=False)
    quantity = Column(Numeric, nullable=True)
    exit_date = Column(DateTime, nullable=True)
    status = Column(Enum(TradeStatus), nullable=False, default=TradeStatus.open)

    now = datetime.utcnow()
    created_at = Column(DateTime, default=now)
    updated_at = Column(DateTime, default=now, onupdate=datetime.utcnow())

    balance_id = Column(Integer, ForeignKey('balances.balance_id'), nullable=False)
    balance = relationship("Balances", back_populates="trades")
    orders = relationship("Orders", back_populates="trade")
    quantfury_position = relationship("QuantfuryPositions", back_populates="trade", uselist=False)

    def __repr__(self):
        return "<Trades(name={}, symbol={}, time_frame={}, date={})>".format(
            self.name, self.symbol, self.time_frame, self.entry_date)

    def is_stop_loss_reached(self, prices):
        # low si no estoy en la misma vela
        # close y low_moment_price si estoy en la misma vela
        log_template = 'func=Trades.is_stop_loss_reached, is_reached={}, stage={}, price={}, stop_loss={}, low={}, '\
                       'low_moment_price={}, trade_id={}, name={}, symbol={}, time_frame={}, entry_date={}, diff={}'
        for date, row in prices.items():
            row['Low'] = Decimal(row['Low'])
            row['Close'] = Decimal(row['Close'])
            entry_date = self.entry_date.strftime("%Y/%m/%d, %H:%M:%S")
            if row['date'] == entry_date:
                if self.low_moment_price > row['Low']:
                    # TODO error in if due to presicion
                    logger.info(log_template.format(self.stop_loss >= row['Low'], 'same candle new low', row['Close'],
                                                    self.stop_loss, row['Low'], self.low_moment_price,
                                                    self.trade_id, self.name, self.symbol, self.time_frame,
                                                    entry_date.replace(',', ''), self.low_moment_price - row['Low']))
                    return self.stop_loss >= row['Low']
                else:
                    logger.info(log_template.format(self.stop_loss >= row['Close'], 'same candle same low',
                                                    row['Close'], self.stop_loss, row['Low'], self.low_moment_price,
                                                    self.trade_id, self.name, self.symbol, self.time_frame,
                                                    entry_date.replace(',', ''), self.low_moment_price - row['Low']))
                    return self.stop_loss >= row['Close']

            else:
                if self.stop_loss >= row['Low']:
                    logger.info(log_template.format('True', 'different candle', row['Close'], self.stop_loss,
                                                    row['Low'], self.low_moment_price,
                                                    self.trade_id, self.name, self.symbol, self.time_frame,
                                                    entry_date.replace(',', ''), self.stop_loss - row['Low']))
                    return True
        logger.info(log_template.format('False', 'last ', row['Close'], self.stop_loss, row['Low'],
                                        self.low_moment_price, self.trade_id, self.name, self.symbol, self.time_frame,
                                        entry_date.replace(',', ''), row['Close'] - self.stop_loss))
        return False

    @classmethod
    def timestamp_to_time_frame(cls, timestamp, time_frame):
        timestamp = timestamp.replace(second=0, microsecond=0)
        if time_frame == '15':
            return timestamp.replace(minute=timestamp.minute//15 * 15)
        elif time_frame == '60':
            return timestamp.replace(minute=0)
        elif time_frame == '240':
            return timestamp.replace(hour=timestamp.hour//4 * 4, minute=0)
        elif time_frame == '1D':
            return timestamp.replace(hour=0, minute=0)
        else:
            raise ValueError('Invalid time_frame')

    def has_open_order(self):
        return self.sell_price is None and self.profit is None

    def cancel_trade(self, client):
        try:
            if self.quantfury_position:
                if self.quantfury_position.status.value == QuantfuryStatus.hold.value:
                    self.sell_price = self.buy_price
                    self.profit = 0
                    self.quantfury_position.calcel(client)
                else:
                    raise ValueError('Trade cannot be cancel because has an open position')
            else:
                self.sell_price = self.buy_price
                self.profit = 0
                self.exit_date = exit_date
                session.commit()
            logger.info('func=Trades.cancel_trade, balance_id={}, trade_id={}'.format(self.balance_id, self.trade_id))
        except Exception as e:
            logger.error('func=Trades.cancel_trade, balance_id={}, trade_id={}, msg={}'.format(
                self.balance_id, self.trade_id, e))

    @property
    def status(self):
        if not self.exit_date:
            if self.buy_price == self.sell_price:
                return 'break-even'
            else:
                return 'open'
        else:
            return 'close'

    @property
    def stop_loss_percentage(self):
        return 1 - min(self.buy_price, self.stop_loss)/max(self.buy_price, self.stop_loss)


class Balances(Base):
    __tablename__ = 'balances'

    balance_id = Column(Integer, primary_key=True, autoincrement=True)
    account_id = Column(Integer, ForeignKey('accounts.account_id'), nullable=False)
    name = Column(Unicode, nullable=False, unique=True)
    symbol = Column(String, nullable=False)
    time_frame = Column(String, nullable=False)
    free = Column(Numeric, nullable=False)
    locked = Column(Numeric, nullable=False)

    stop_loss_percentage = Column(Numeric, nullable=False)
    profit_percentage = Column(Numeric, nullable=False)
    trade_in_trend = Column(Enum(TradeInTrend), nullable=False, default=False)
    second_scale_from_database = Column(Boolean, nullable=False, default=True)
    activate_alert = Column(Boolean, nullable=False, default=False)

    created_at = Column(DateTime(timezone=False), default=datetime.utcnow())
    updated_at = Column(DateTime(timezone=False), default=datetime.utcnow(), onupdate=datetime.utcnow())
    account = relationship("Accounts", back_populates="balances")
    trades = relationship("Trades", back_populates="balance", order_by="desc(Trades.updated_at)")

    @classmethod
    def balances_by_strategy(cls, symbol, time_frame):
        records = session.query(Balances).filter(
            Balances.symbol == symbol, Balances.time_frame == time_frame,
            Balances.profit_percentage != 0).all()
        return records

    @classmethod
    def balances_by_stop_loss_percentage(cls, symbol, time_frame, stop_loss_percentage):
        records = session.query(Balances).filter(
            Balances.symbol == symbol, Balances.time_frame == time_frame, Balances.stop_loss_percentage == stop_loss_percentage,
            Balances.profit_percentage != 0).all()
        return records

    def map_time_frame_to_minutes(cls, time_frame):
        if time_frame.isdigit():
            return int(time_frame)
        elif time_frame == '1D':
            return 1440
        elif time_frame == '1W':
            return 10080
        else:
            raise ValueError('Invalid time_frame')

    @property
    def delay_to_open_a_position(self):
        return self.map_time_frame_to_minutes(self.time_frame)

    @property
    def latest_trade(self):
        if self.trades:
            return self.trades[0]

    def __repr__(self):
        return "<Balances(balance_id={}, symbol={}, time_frame={}, free={})>".format(
            self.balance_id, self.symbol, self.time_frame, self.free)

    @property
    def symbol_metadata(self):
        symbol = self.symbol.split(':')[1] if ':' in self.symbol else self.symbol
        return session.query(SymbolMetadata).filter(
            (SymbolMetadata.symbol == symbol) | (SymbolMetadata.currency_pair == symbol)).first()


class Accounts(Base):
    __tablename__ = 'accounts'
    if not cypher_key:
        raise ValueError('CYPHER_KEY environment variable is required')
    account_id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(StringEncryptedType(String, cypher_key), unique=True, nullable=False)
    api_key = Column(StringEncryptedType(String, cypher_key), unique=True, nullable=False)
    secret_key = Column(StringEncryptedType(String, cypher_key), nullable=False)
    token = Column(StringEncryptedType(String, cypher_key), nullable=True)
    broker_name = Column(String, nullable=False)

    created_at = Column(DateTime, default=datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.utcnow(), onupdate=datetime.utcnow())
    #__table_args__ = (UniqueConstraint('api_key', name='api_key_unique'),)
    balances = relationship("Balances", back_populates="account")

    def masked_api_key(self):
        prefix = len(self.name)//3 * '*'
        return "{}{}{}".format(prefix, self.name[len(self.name)//3:len(self.name)//3*2], prefix)

    def __repr__(self):
        return "<Accounts(account_id={}, name={}, api_key={})>".format(
            self.account_id, self.name, self.api_key)


@event.listens_for(QuantfuryPositions.position_open_price, 'set')
def sync_open_price(target, value, old_value, initiator):
    if target.trade:
        target.trade.buy_price = value


@event.listens_for(QuantfuryPositions.close_price, 'set')
def sync_close_price(target, value, old_value, initiator):
    if target.trade:
        target.trade.sell_price = value


@event.listens_for(QuantfuryPositions.profit_and_loss_system, 'set')
def sync_profit_and_loss_system(target, value, old_value, initiator):
    if target.trade:
        target.trade.profit = target.trade.amount + Decimal(str(value))


@event.listens_for(QuantfuryPositions.limit_stop_order, 'set')
def sync_stop_order(target, value, old_value, initiator):
    if target.trade:
        target.trade.stop_loss = value



import logging
import decimal
from broker_clients import session
from broker_clients import Balances, Trades

MINIMUM_TRADE_AMOUNT = 0
logger = logging.getLogger(__name__)


class BrokerEmulator:
    def __init__(self, balance_name=None, balance_id=None):
        if balance_name:
            self.balance_name = balance_name
        if balance_id:
            self.balance_id = balance_id

        self.balance = self.get_balance(balance_name, balance_id)

    def get_balance(self, balance_name=None, balance_id=None):
        if balance_name is None and balance_id is None:
            raise ValueError('No balance id or name provided')
        balance = None

        if balance_name:
            balance = session.query(Balances).filter_by(name=balance_name).first()
        if balance is None:
            balance = session.query(Balances).filter_by(balance_id=balance_id).first()
            if balance is None:
                raise ValueError('No Balance found with name: {} or id: {}'.format(balance_name, balance_id))
            else:
                self.balance_name = balance.name
        else:
            self.balance_id = balance.balance_id
        return balance

    def has_open_order(self):
        return self.count_opened_trades() == 1

    def query_open_trades(self):
        return session.query(Trades).filter_by(name=self.balance_name, sell_price=None, profit=None)

    def count_opened_trades(self):
        return self.query_open_trades().count()

    @ property
    def stop_loss(self):
        return self.trade.stop_loss

    @ stop_loss.setter
    def stop_loss(self, value):
        self.trade.stop_loss = value
        session.commit()

    @property
    def trade(self):
        return self.query_open_trades().first()

    @property
    def trade_id(self):
        return self.trade.trade_id

    @property
    def trade_budget(self):
        return self.balance.free

    def market(self, side, amount, stop_loss=None, stop_loss_percentage=None, emulated_time=None):
        if side not in ['BUY', 'SELL'] or amount <= MINIMUM_TRADE_AMOUNT:
            raise ValueError('Invalid side: {} or amount: {} in market order'.format(side, amount))
        if stop_loss is None and stop_loss_percentage is None:
            raise ValueError('Stop loss must be set')
        if self.count_opened_trades() >= 1:
            raise ValueError('There are {} opened trades'.format(side))
        close_price = decimal.Decimal(emulated_time['Close'])
        calculated_stop_loss = None
        if side == 'BUY':
            if not stop_loss:
                calculated_stop_loss = close_price - (close_price * stop_loss_percentage)

            trade = Trades(name=self.balance_name,
                           symbol=self.balance.symbol,
                           time_frame=self.balance.time_frame,
                           entry_date=emulated_time['date'],
                           amount=amount,
                           stop_loss=calculated_stop_loss,
                           buy_price=close_price,
                           sell_price=None,
                           profit=None,
                           side=side,
                           low_moment_price=emulated_time['Low'],
                           status=TradeStatus.open.value,
                           balance_id=self.balance_id)
            session.add(trade)
            session.commit()

    def take_profit(self, row=None):
        close_price = decimal.Decimal(row['Close'])
        delta = close_price - self.trade.buy_price
        profit_percentage = delta / close_price
        profit_amount = self.trade.amount * profit_percentage
        self.set_sell_price(close_price, self.trade.amount + profit_amount, row)
        return {
            "profit_percentage": profit_percentage,
            "profit_amount": profit_amount
        }

    def set_sell_price(self, sell_price, profit, row):
        current_trade = self.trade
        current_trade.sell_price = sell_price
        current_trade.profit = self.balance.symbol_metadata.fix_price_precision(profit)
        current_trade.exit_date = row['date']
        self.balance.free = current_trade.profit
        session.commit()


if __name__ == '__main__':
    balance_name = ''
    broker = BrokerEmulator(balance_name)
    side, amount, stop_loss, stop_loss_percentage, emulated_time = 'BUY', 100, 145, None, {
        "Close": 150, "date": "2021/01/14, 14:30:00"}
    broker.market(side, amount, stop_loss=stop_loss, emulated_time=emulated_time)



"""Change trade_in_trend type

Revision ID: e84250567e04
Revises: 5a0ada3a67ed
Create Date: 2021-06-25 21:13:10.517179

"""
from alembic import op
import os
import json
from sqlalchemy import MetaData, Table


def get_module_path(module_name):
    import re
    return re.sub(module_name, '', __file__)

# revision identifiers, used by Alembic.
revision = 'e84250567e04'
down_revision = '5a0ada3a67ed'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("""
    CREATE TYPE tradeintrend AS ENUM (
    'none', 'first_upper_scale', 'second_upper_scale',
    'both_upper_scales', 'any_upper_scales', 'second_any_upper_scales');""")

    op.execute("""
    ALTER TABLE balances
        ALTER COLUMN trade_in_trend TYPE tradeintrend USING (
            CASE trade_in_trend
            WHEN TRUE  THEN 'any_upper_scales'::tradeintrend
            WHEN FALSE THEN 'none'::tradeintrend
            END);""")

    module_path = get_module_path(r'/alembic/versions.+')
    path = os.path.join(module_path, 'data/balances_v2.json')
    f = open(path)
    balances_initial_data = json.load(f)
    meta = MetaData(bind=op.get_bind())
    meta.reflect(only=('balances',))
    table = Table('balances', meta)
    op.bulk_insert(table, balances_initial_data)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    #f = open('data/balances_v2.json')
    #balances_initial_data = json.load(f)
    #balances_names = [i['name'] for i in balances_initial_data]
    #session.query(Trades).filter(Trades.name.in_(balances_names)).delete(synchronize_session='fetch')
    #session.query(Balances).filter(Balances.name.in_(balances_names)).delete(synchronize_session='fetch')
    #session.commit()
    op.execute("""
    ALTER TABLE balances
        ALTER COLUMN trade_in_trend TYPE BOOLEAN USING (
            CASE trade_in_trend
            WHEN 'first_upper_scale' THEN TRUE
            WHEN 'second_upper_scale' THEN TRUE
            WHEN 'both_upper_scales' THEN TRUE
            WHEN 'any_upper_scales' THEN TRUE
            WHEN 'second_any_upper_scales' THEN TRUE
            WHEN 'none' THEN FALSE
            END);""")
    op.execute("DROP TYPE tradeintrend;")
    # ### end Alembic commands ###



import json
from websocket import create_connection
import requests
import urllib
import logging

logger = logging.getLogger(__name__)

SYMBOL_SWITCHER = {
    "BTCUSDT": "BTC/USDT"
}

POSITION_TYPES = {
    "long": 1,
    "short": 2
}

ORDER_TYPES = {
    "stop_order": 0,
    "target_order": 1,
    "limit_order": 2,
}

REVERTED_POSITION_TYPES = {v: k for k, v in POSITION_TYPES.items()}
REVERTED_ORDER_TYPES = {v: k for k, v in ORDER_TYPES.items()}

DOMAIN = 'e1.quantfury.com'
WEBSOCKET_DOMAIN = 'a1.quantfury.com'
VERSION = 'v7'
NEGOTIATE_URL = "https://{}/signalr/negotiate?clientProtocol=2.1&_=1626042979667".format(WEBSOCKET_DOMAIN)
WEBSOCKET_URL = 'wss://{domain}/signalr/connect'\
                '?transport=webSockets&clientProtocol=2.1&connectionToken={{}}&tid=0'.format(domain=WEBSOCKET_DOMAIN)

common_headers = {
    "Host": "e1.quantfury.com",
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Accept": "application/json, text/plain, */*",
    "Accept-Language": "en-US,en;q=0.5",
    "Accept-Encoding": "deflate",
    "Content-Type": "application/json;charset=utf-8",
    "Origin": "https://trading.quantfury.com",
    "Connection": "keep-alive",
    "Referer": "https://trading.quantfury.com/"
}


class Utils:

    @classmethod
    def reformat_symbol(cls, symbol):
        if type(symbol) is not str:
            raise ValueError('Invalid symbol Type')
        if symbol in SYMBOL_SWITCHER:
            return SYMBOL_SWITCHER[symbol]
        elif symbol.find('/') < 0:
            logger.warning('func=Utils.reformat_symbol, symbol={}, msg=Symbol should have Slash'.format(symbol, ))
        return symbol

    @classmethod
    def get_connection_token(cls):
        status_code, reason = None, None
        payload = {}
        headers = {}
        try:
            raw_response = requests.request("GET", NEGOTIATE_URL, headers=headers, data=payload)

            status_code, reason = raw_response.status_code, raw_response.reason
            response = raw_response.json()

            connection_token = response['ConnectionToken']
            connection_token = urllib.parse.quote(connection_token)

            return connection_token
        except Exception as e:
            logger.critical('func=Utils.get_connection_token, status_code={}, reason={}, msg={}'.format(
                status_code, reason, e))
            raise e

    @classmethod
    def get_price_id(cls, symbol):
        return cls.get_latest_price_and_id(symbol)[1]

    @classmethod
    def get_latest_price(cls, symbol):
        return cls.get_latest_price_and_id(symbol)[0]

    @classmethod
    def get_latest_price_and_id(cls, symbol):
        try:
            symbol = cls.reformat_symbol(symbol)
            connection_token = cls.get_connection_token()

            headers = json.dumps({

            })
            ws = create_connection(
                WEBSOCKET_URL.format(
                    connection_token), headers=headers)

            ws.send(json.dumps({"H": "priceshub", "M": "SubscribeToPriceStreams", "A": [[symbol]], "I": 0}))

            for _ in range(10000):
                result = ws.recv()
                if result.find('"shortName":"{}"'.format(symbol)) > 0 and result.find('"id":"') > 0:
                    price_data = json.loads(result)
                    return price_data['M'][0]['A'][0]['price'], price_data['M'][0]['A'][0]['id']
        except Exception as e:
            logger.critical('func=Utils.get_latest_price_and_id, symbol={}, msg={}'.format(symbol, e))
            raise e

    @classmethod
    def get_test_limit_stop_target_prices(cls, symbol):
        buy_price = Utils.get_latest_price(symbol)
        buy_price = buy_price - (buy_price % 100) - 50
        stop_price, target_price = buy_price - 1000, buy_price + 1000
        return buy_price, stop_price, target_price

    @classmethod
    def get_device_id(cls):
        return "ead2ce58-63e1-414d-bda7-104f0069edd0"


class Quantfury:
    def __init__(self, token):
        self.token = token

    def create_OCO_long_order(self, symbol, quote_order_quantity, buy_price, stop_price, target_price, position_type):
        """
        Create OCO long order
        Parameters
        ----------
        symbol
        quote_order_quantity
        buy_price
        stop_price
        target_price
        position_type

        Returns
        -------

        """
        url = 'https://{domain}/{version}/limitOrder/create'.format(domain=DOMAIN, version=VERSION)
        status_code, reason, error_code, error_message, stage = None, None, None, None, 'API.first_attempt'
        try:
            latest_price, price_id = Utils.get_latest_price_and_id(symbol)
            body = {
                "priceId": price_id,
                "price": buy_price,
                "value": {
                    "amountSystem": quote_order_quantity
                },
                "stopOrder": stop_price,
                "targetOrder": target_price,
                "positionType": POSITION_TYPES[position_type],
                "executionType": 0,
                "shortName": symbol
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason or None
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            if error_code == 'InvalidLimitOrderPrice':
                stage = 'API.second_attempt'
                body['price'] = latest_price
                raw_response = requests.post(url, json=body, headers=headers)

                status_code, reason = raw_response.status_code, raw_response.reason or None
                json_result = raw_response.json()
                error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.create_OCO_long_order, symbol={}, quote_order_quantity={}, buy_price={}, '
                'stop_price={}, target_price={}, position_type={}, status_code={}, reason={}, error_code={}, '
                'error_message={}, stage={}, msg={}'.format(
                    symbol, quote_order_quantity, buy_price, stop_price, target_price, position_type, status_code,
                    reason, error_code, error_message, stage, e))
            raise e

    def create_limit_long_order(self, symbol, quote_order_quantity, buy_price, position_type):
        """
        Create long limit order
        Parameters
        ----------
        symbol
        quote_order_quantity
        buy_price
        position_type

        Returns
        -------

        """
        url = 'https://{domain}/{version}/limitOrder/create'.format(domain=DOMAIN, version=VERSION)
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            body = {
                "priceId": Utils.get_price_id(symbol),
                "price": buy_price,
                "executionType": 0,
                "value": {
                    "amountSystem": quote_order_quantity
                },
                "positionType": POSITION_TYPES[position_type],
                "shortName": symbol
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)


            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.create_limit_long_order, symbol={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    symbol, status_code, reason, error_code, error_message, e))
            raise e

    def create_market_order(self, symbol, quote_order_quantity, position_type):
        """
        Create market order
        Parameters
        ----------
        symbol
        quote_order_quantity
        position_type

        Returns
        -------

        """
        url = 'https://{domain}/{version}/positions/open'.format(domain=DOMAIN, version=VERSION)
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            body = {
                "shortName": symbol,
                "positionType": POSITION_TYPES[position_type],
                "priceId": Utils.get_price_id(symbol),
                "targetOrders": [],
                "stopOrders": [],
                "value": {
                    "amountSystem": quote_order_quantity
                }
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)


            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.create_market_order, symbol={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    symbol, status_code, reason, error_code, error_message, e))
            raise e

    def update_position(self, api_order_id, symbol, price_value, quote_order_quantity, order_type):
        """
        Update a position
        Parameters
        ----------
        position_id : str
            Position ID
        symbol : str
            Asset symbol
        price_value : float
        quote_order_quantity : int
            Quantity in usdt
        order_type : str
            Order type stop_order: 0, target_order: 1, limit_order: 2

        Returns
        -------
        dict
            Data of the position
        """
        url = 'https://{domain}/{version}/reduceOrders/update'.format(domain=DOMAIN, version=VERSION)
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            if order_type not in ORDER_TYPES:
                raise ValueError('Invalid order_type <{}> must be {}'.format(order_type, ';'.join(ORDER_TYPES.keys())))
            body = {
                "price": price_value,
                "id": api_order_id,
                "value": {
                    "amountSystem": quote_order_quantity
                }
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.update_position, symbol={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    symbol, status_code, reason, error_code, error_message, e))
            raise e

    def create_position(self, position_id, symbol, price_value, quote_order_quantity, order_type):
        """
        Update a position
        Parameters
        ----------
        position_id : str
            Position ID
        symbol : str
            Asset symbol
        price_value : float
        quote_order_quantity : int
            Quantity in usdt
        order_type : str
            Order type stop_order: 0, target_order: 1, limit_order: 2

        Returns
        -------
        dict
            Data of the position
        """
        url = 'https://{domain}/{version}/reduceOrders/create'.format(domain=DOMAIN, version=VERSION)
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            if order_type not in ORDER_TYPES:
                raise ValueError('Invalid order_type <{}> must be {}'.format(order_type, ';'.join(ORDER_TYPES.keys())))
            body = {
                "price": price_value,
                "tradingPositionId": position_id,
                "orderType": ORDER_TYPES[order_type],
                "value": {
                    "amountSystem": quote_order_quantity
                }
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.create_position, symbol={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    symbol, status_code, reason, error_code, error_message, e))
            raise e

    def update_order(self, order_id, symbol, buy_price, stop_price, target_price, quote_order_quantity, order_type):
        """
        Update an order
        Parameters
        ----------
        order_id : str
            Order id
        symbol : str
            Asset symbol
        buy_price : float
            Buy price
        stop_price : float
            Stop price
        target_price : float
            target price
        quote_order_quantity : int
            Quantity in Dollars
        order_type : str
            Order type stop_order: 0, target_order: 1, limit_order: 2
        Returns
        -------
        dict
            Data of order

        """
        url = 'https://{domain}/{version}/limitOrder/update'.format(domain=DOMAIN, version=VERSION)
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            if order_type not in ORDER_TYPES:
                raise ValueError('Invalid order_type <{}> must be {}'.format(order_type, ';'.join(ORDER_TYPES.keys())))
            body = {
                "priceId": Utils.get_price_id(symbol),
                "price": buy_price,
                "executionType": 0,
                "value": {
                    "amountSystem": quote_order_quantity
                },
                "stopOrder": stop_price,
                "targetOrder": target_price,
                "id": order_id
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.update_order, order_id={}, symbol={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    order_id, symbol, status_code, reason, error_code, error_message, e))
            raise e

    @classmethod
    def login(cls, username, password):
        """

        Parameters
        ----------
        username : str
            Username of a Quantfury Account
        password : str
            Password of a Quantfury Account
        Returns
        -------
        dict
            Dict with token field
        """
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            url = 'https://{domain}/{version}/auth/login'.format(domain=DOMAIN, version=VERSION)
            country_code = 52
            country_code_iso = 'MX'
            device_id = Utils.get_device_id()
            body = {
                "password": password,
                "phoneNumber": username,
                "countryCode": country_code,
                "countryCodeIso": country_code_iso,
                "deviceId": device_id
            }
            raw_response = requests.post(url, data=json.dumps(body), headers=common_headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            n = len(username)//2
            masked_username = ("*"*n)+username[n:]
            logger.critical(
                'func=Quantfury.login, user={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    masked_username, status_code, reason, error_code, error_message, e))
            raise e

    def check_credentials(self):
        """
        Check if token is valid
        Returns
        -------
        dict
            True if token works, False otherwise
        """
        try:
            orders = self.get_account_info()
            return 'data' in orders
        except Exception as _:
            return False

    def close_order(self, order_id):
        """
        Close order
        Parameters
        ----------
        order_id

        Returns
        -------

        """
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            url = 'https://{domain}/{version}/limitOrder/cancel'.format(domain=DOMAIN, version=VERSION)
            body = {
                "tradingPositionId": order_id,
                "priceId": Utils.get_price_id(symbol)
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.cancel, order_id={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    order_id, status_code, reason, error_code, error_message, e))
            raise e

    def cancel(self, order_id):
        """
        Cancel order
        Parameters
        ----------
        order_id : str
            Order ID to cancel
        Returns
        -------
        dict
            Dict with code of success or fail

        """
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            url = 'https://{domain}/{version}/limitOrder/cancel'.format(domain=DOMAIN, version=VERSION)
            body = {
                "id": order_id
            }
            headers = self.get_headers()
            raw_response = requests.post(url, json=body, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.cancel, order_id={}, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    order_id, status_code, reason, error_code, error_message, e))
            raise e

    def get_account_info(self):
        """
        Get account information, open positions and active limit orders
        Returns
        -------
        dict
            Dict with fields: open_positions and active_limit_orders
        """
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            url = 'https://{domain}/{version}/session'.format(domain=DOMAIN, version=VERSION)

            headers = self.get_headers()
            raw_response = requests.get(url, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.get_account_info, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    status_code, reason, error_code, error_message, e))
            raise e

    def active_trade(self, symbol):
        """
        Check if there is an open trade
        Parameters
        ----------
        symbol : str
            Asset symbol

        Returns
        -------
        bool
            True if there is an open position for the symbol

        """
        orders = self.get_account_info()
        active_limit_orders = orders['data']['activeLimitOrders']
        open_positions = orders['data']['openPositions']
        if len([i for i in active_limit_orders if i['shortName'] == symbol]) > 0:
            return active_limit_orders[0]['id']
        elif len([i for i in open_positions if i['shortName'] == symbol]) > 0:
            return open_positions[0]['id']
        return None

    def active_order_in_position(self, symbol, order_type):
        """
        Check if there is an active order in an open position
        Parameters
        ----------
        symbol : str
            Asset symbol
        order_type : str
            Order type

        Returns
        -------
        bool
            order_api_id

        """
        try:
            if order_type in ORDER_TYPES.keys():
                order_type = ORDER_TYPES[order_type]
            elif order_type not in ORDER_TYPES.values():
                raise ValueError('Invalid order type')

            orders = self.get_account_info()
            open_positions = orders['data']['openPositions']
            for open_position in open_positions:
                if open_position['shortName'] == symbol and open_position['stopOrders']:
                    if len(open_position['stopOrders']) > 1:
                        logger.warning('func=Quantfury.active_order_in_position, symbol={}, order_type={}, '
                                       'msg=Many declared stop orders'.format(symbol, order_type))
                    for order in open_position['stopOrders']:
                        if order['orderType'] == order_type:
                            return order['id']
                if open_position['shortName'] == symbol and open_position['targetOrders']:
                    if len(open_position['targetOrders']) > 1:
                        logger.warning('func=Quantfury.active_order_in_position, symbol={}, order_type={}, '
                                       'msg=Many declared target orders'.format(symbol, order_type))
                    for order in open_position['targetOrders']:
                        if order['orderType'] == order_type:
                            return order['id']
            return None
        except Exception as e:
            logger.critical('func=Quantfury.active_order_in_position, symbol={}, order_type={}, msg={}'.format(
                symbol, order_type, e))

    def upsert_position(self, api_position_id, symbol, order_type, target_price, limit_amount_system):
        api_order_id = self.active_order_in_position(symbol, order_type)
        if api_order_id:
            order_json_result = self.update_position(
                api_order_id, symbol, target_price,
                limit_amount_system, order_type)
        else:
            order_json_result = self.create_position(
                api_position_id, symbol, target_price,
                limit_amount_system, order_type)
        return order_json_result

    def list_closed_positions(self):
        """
        List close position in the account
        Returns
        -------
        list
            List with one dict peer close position
        """
        status_code, reason, error_code, error_message = None, None, None, None
        try:
            url = 'https://{domain}/{version}/tradingHistory/closedAndReducedPositions'.format(
                domain=DOMAIN, version=VERSION)

            headers = self.get_headers()
            raw_response = requests.post(url, json={}, headers=headers)

            status_code, reason = raw_response.status_code, raw_response.reason
            json_result = raw_response.json()
            error_code, error_message = json_result.get('code', None), json_result.get('error', None)

            return json_result
        except Exception as e:
            logger.critical(
                'func=Quantfury.list_closed_positions, '
                'status_code={}, reason={}, error_code={}, error_message={}, msg={}'.format(
                    status_code, reason, error_code, error_message, e))
            raise e

    def get_account_balance(self):
        session_data = self.get_account_info()
        account_info = session_data['data']
        return account_info['balanceAccount'] + account_info['pnlAccountClosePositions'] + \
            account_info['pnlAccountDividends']

    def get_headers(self):
        return {**common_headers, **{
            "Authorization": "Bearer {}".format(self.token)}}

    @classmethod
    def filter_order(cls, account_info_response, order_id):
        orders = [i for i in account_info_response['data']['activeLimitOrders'] if order_id == i['id']]
        if orders:
            return orders[0]

    @classmethod
    def calculate_open_position_profits(cls, orders):
        profits = {}
        if not ('data' in orders and 'openPositions' in orders['data']):
            return {}
        for position in orders["data"]["openPositions"]:
            current_price = Utils.get_latest_price(symbol)
            min_price_tmp = min(position["openPrice"], current_price)
            profits[position['shortName']] = min_price_tmp / max(position["openPrice"], current_price)
        return profits

    def get_open_position_profits(self):
        orders = self.get_account_info()
        return self.calculate_open_position_profits(orders)


def test_oco_order(username, password, symbol, token):
    quote_order_quantity = 10
    position_type = 'long'
    buy_price = Utils.get_latest_price(symbol)
    buy_price = buy_price - buy_price % 100

    stop_price, target_price = buy_price - 1000, buy_price + 1000
    if token:
        quantfury = Quantfury(token)
    else:
        login_result = Quantfury.login(username, password)
        quantfury = Quantfury(login_result['data']['token'])

    oco_response = quantfury.create_OCO_long_order(
        symbol, quote_order_quantity, buy_price, stop_price, target_price, position_type)
    order_id = oco_response['data']['id']
    logger.info(
        'func=test_oco_order, stage=open_order, order_id={}, buy_price={}, stop_price={}, target_price={}'.format(
            order_id, buy_price, stop_price, target_price))
    print(order_id)
    account_info_response = quantfury.get_account_info()
    if len(account_info_response['data']['activeLimitOrders']) == 0:
        raise ValueError('No active orders')

    order_id_listed = account_info_response['data']['activeLimitOrders'][0]['id']
    assert order_id == order_id_listed

    price_value = buy_price - 200
    order_type = 'LIMIT_ORDER'
    update_order_response = quantfury.update_order(
        order_id, symbol, price_value, stop_price, target_price, quote_order_quantity, order_type)

    assert 'data' in update_order_response

    response_buy_price = update_order_response['data']['price']
    response_stop_price = update_order_response['data']['stopOrder']
    response_target_price = update_order_response['data']['targetOrder']
    update_order_response_id = update_order_response['data']['id']
    logger.info(
        'func=test_oco_order, msg=update_order, order_id={}, buy_price={}, stop_price={}, target_price={}'.format(
            update_order_response_id, response_buy_price, response_stop_price, response_target_price))
    assert price_value == response_buy_price
    assert stop_price == response_stop_price
    assert target_price == response_target_price
    assert order_id == update_order_response_id

    quantfury.cancel(order_id_listed)

    account_info_response = quantfury.get_account_info()
    order = quantfury.filter_order(account_info_response, order_id)
    assert order is None
    logger.info('func=test_oco_order, msg=cancel_order, order_id={}'.format(update_order_response_id))


if __name__ == '__main__':
    import os
    from dotenv import load_dotenv
    load_dotenv()

    username = os.getenv("QUANTFURY_USERNAME")
    password = os.getenv("QUANTFURY_PASSWORD")
    token = os.getenv("QUANTFURY_TOKEN")
    symbol = 'BTC/USDT'

    test_oco_order(username, password, symbol, token)





"""Add column status in Trade

Revision ID: 33c8546d11e2
Revises: 86498a37f807
Create Date: 2022-08-13 18:55:08.104304

"""
from alembic import op
import sqlalchemy as sa
import enum
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = '33c8546d11e2'
down_revision = '86498a37f807'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    class TradeStatus(enum.Enum):
        hold = 'hold'
        open = 'open'
        close = 'close'
        cancel = 'cancel'
        target = 'target'

    trade_status = postgresql.ENUM(TradeStatus, name="tradestatus")
    trade_status.create(op.get_bind(), checkfirst=True)

    op.add_column('trades', sa.Column('status', trade_status, nullable=True, default=TradeStatus.open.value))
    op.execute("UPDATE trades SET status = 'close'")
    op.execute("UPDATE trades SET status = 'open' WHERE sell_price is NULL")
    op.alter_column('trades', 'status', nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_column('trades', 'status')
    op.execute("DROP TYPE tradestatus;")

    # ### end Alembic commands ###



import logging
from broker_clients.quantfury_proxy.broker.client import Utils
logger = logging.getLogger(__name__)


class ExtendedAccount:

    def __init__(self, name, symbol):
        pass

    def market(self, trade_id, symbol, side, quote_order_quantity):
        try:
            order = Orders.register_order(self.client, trade_id, symbol=symbol, side=side,
                                          quoteOrderQty=quote_order_quantity, type=self.client.ORDER_TYPE_MARKET)
            return order
        except Exception as e:
            logger.error('func=ExtendedAccount.market, account_name={}, trade_id={}, symbol={},'
                         'side={}, quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quote_order_quantity, e))
            raise e

    def market_quantity(self, trade_id, symbol, side, quantity):
        try:
            pass
        except Exception as e:
            logger.error('func=ExtendedAccount.market, account_name={}, trade_id={}, symbol={},'
                         'side={}, quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quantity, e))
            raise e

    def limit_buy(self, trade_id, symbol, quote_order_quantity, price):
        try:
            pass
        except Exception as e:
            logger.error('func=ExtendedAccount.limit_buy, account_name={}, trade_id={}, symbol={}, '
                         'quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, quote_order_quantity, e))
            raise e

    def limit_sell(self, trade_id, symbol, quote_order_quantity, price):
        try:
            pass
        except Exception as e:
            logger.error('func=ExtendedAccount.limit_sell, account_name={}, trade_id={}, symbol={},'
                         'side={}, quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quote_order_quantity, e))
            raise e

    def take_profit_limit(self, symbol, side, quote_order_quantity, price, stop_price):
        try:
            pass

        except Exception as e:
            logger.error('func=ExtendedAccount.take_profit_limit, symbol={}, quote_order_quantity={}, price={}, '
                         'msg={}'.format(symbol, quote_order_quantity, price, e))
            raise e

    def market_with_stop_loss(self, balance_name, side):
        stage = 'API.read_price'
        try:
            pass
        except Exception as e:
            logger.critical('func=ExtendedAccount.market_with_stop_loss, account_name={}, symbol={}, stage={}, '
                            'msg={}'.format(self.name, self.symbol, stage, e))


    def OCO_buy(self, trade_id, symbol, quote_order_quantity, buy_price, target_price, stop_price):
        try:
            symbol = Utils.reformat_symbol(symbol)

        except Exception as e:
            logger.error('func=ExtendedAccount.limit_buy, account_name={}, trade_id={}, symbol={}, '
                         'quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, quote_order_quantity, e))
            raise e


"""Add exit date in trades table

Revision ID: 14f9fd83af6a
Revises: b1f321f39713
Create Date: 2021-05-12 20:16:21.503507

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '14f9fd83af6a'
down_revision = 'b1f321f39713'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('trades', sa.Column('exit_date', sa.DateTime(), nullable=True))
    op.alter_column('trades', 'date', nullable=False, new_column_name='entry_date')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('trades', 'entry_date', nullable=False, new_column_name='date')
    op.drop_column('trades', 'exit_date')
    # ### end Alembic commands ###



from setuptools import setup, find_packages


setup(
    name='broker_clients',
    version='0.3.7',
    license='UNLICENSED',
    author="Oswaldo Cruz",
    author_email='oswaldo_cs_94@hotmail.com',
    url='https://github.com/gmyrianthous/example-publish-pypi',
    keywords='example project',
    entry_points={
        'console_scripts': ['account_manager=broker_clients.binance_proxy.broker.account_manager:command'],
    },
    packages=['broker_clients'],
    package_data={
        '': [
            'alembic/*',
            'data/*',
            'alembic/versions/*',
            'emulator_proxy/*/*.py',
            'emulator_proxy/*.py',
            'quantfury_proxy/*/*.py',
            'quantfury_proxy/*.py',
            'binance_proxy/*/*.py',
            'binance_proxy/*.py',
            'alembic.ini'
        ],
    }
)